/*
    Fichier trtligne.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence.fr

    Traductions des droits d'utilisation dans les fichiers :
    licence.de , licence.en , licence.es , licence.it
    licence.nl , licence.pt , licence.eo , licence.eo-utf


    Bibliothque de fonctions permettant de :

    - mmoriser le mode d'encodage et le jeu de caractres du message
    - convertir les accents et caractres spciaux cods sur 7 bits
      ou/et utilisant un jeu de caractres non ISO-8859 dans le message
    - rcuprer un nom de fichier joint
*/

#include <stdio.h>
#include <string.h>
#include "buflect.h"
#include "encodage.h"
#include "trtligne.h"
#include "messages.h"    // pour fonction  conv_iso_utf8 (...)
#include "base64.h"


/* prototypes de fonctions locales  ce source */

void sup_quoted ();
void conv_isomac ();
void conv_utf8_iso ();


/* variable globale au source */

static int charset_texte;      // jeu de caractres accentus du texte



/* retourne la valeur du champ Content-Type */

int recup_ctype ()
{
    int i;   // numro de caractre dans buf_lect


    // se positionner aprs le nom de champ
    i = 13;

    // sauter les blancs
    while (buf_lect [i] == ' ')
        i++;

    // dtection du type de contenu
    if (tolower (buf_lect [i]) == 't')
    {
        // discrimination entre text/plain et text/html
        if (tolower (buf_lect [i + 5]) == 'p')
            return (TextPlain);
        else if (tolower (buf_lect [i + 5]) == 'h')
            return (TextHtml);
        else
            return (AutreType);
    }
    else if (tolower (buf_lect [i]) == 'm')
    {
        // type message ou type multipart
        if (tolower (buf_lect [i + 1]) == 'e')
            // message/rfc822 (trait comme section texte)
            return (Mesrfc822);

        // discrimination entre les diffrents mode multipart
        else if (tolower (buf_lect [i + 10]) == 'a')
            // multipart/alternative
            return (MultipAlter);

        else if (tolower (buf_lect [i + 10]) == 'm')
            // multipart/mixed
            return (MultipMixed);

        else if (tolower (buf_lect [i + 10]) == 'r')
        {
            if (tolower (buf_lect [i + 12]) == 'l')
                // multipart/related
                return (MultipRel);
            else if (tolower (buf_lect [i + 12]) == 'p')
                // multipart/report
                return (MultipRep);
        }
        else
            return (Multipart);
    }
    else
        return (AutreType);
}



/* mmorise dans  encodage_texte  la valeur
   du champ Content-Transfer-Encoding */

void mem_encodage ()
{
    int i;   // numro de caractre dans buf_lect


    // se positionner aprs le nom de champ
    i = 26;

    // sauter les blancs
    while (buf_lect [i] == ' ')
        i++;

    // dtection de l'encodage (simplifie, on analyse un ou deux caractre)
    if (tolower (buf_lect [i]) == 'q')
        encodage_texte = QuotedPrint;
    // l'encodage peu frquent "binary" ne doit pas tre trait comme "base64"
    else if (tolower (buf_lect [i]) == 'b' && tolower (buf_lect [i+1]) == 'a')
        encodage_texte = Base64;
    else
        encodage_texte = xBits;
}



/* mmorise dans  charset_texte  le jeu de
   caractres utilis dans le texte du mail
   retourne 1 si champ charset trouv dans la ligne, 0 sinon */

int mem_charset ()
{
    int i;  // compteur


    // initialisation
    i = 1;

    // chercher le mot  charset  dans la ligne et sortir s'il n'y est pas
    while (tolower (buf_lect [i]) != 'c' || tolower (buf_lect [i+1]) != 'h')
        if (! buf_lect [++i])
            return (0);

    // se positionner aprs le caractre  =
    while (buf_lect [i++] != '=')
        if (! buf_lect [i])
            return (0);

    // sauter galement le  "  s'il y est
    if (buf_lect [i] == '"')
        i++;

    // dtection du charset (simplifie, on analyse un caractre)
    if (tolower (buf_lect [i]) == 'i')
    {
        // jeu de caractre ISO
        charset_texte = Iso8859;

        // vrifier si variante Mac
        while (tolower (buf_lect [i]) != 'x' || buf_lect [i+1] != '-'
                                    || tolower (buf_lect [i+2]) != 'm')
            if (! buf_lect [++i])
                return (1);

        // variante Mac trouve
        charset_texte = IsoMac;
    }
    else if (tolower (buf_lect [i]) == 'u')
        charset_texte = Utf8;
    else
        charset_texte = xCharset;

    // on a trouv le charset
    return (1);
}



/* convertit la ligne lue en interprtant les caractres spciaux */

void majligne ()
{
    int i;  // compteur


    // suppression du . ventuel en premire colonne
    if (buf_lect [0] == '.')
    {
        i = 0;

        do
            buf_lect [i] = buf_lect [i + 1];
        while (buf_lect [i++] != '\0');
    }

    // suppression de l'encodage quoted-printable
    if (encodage_texte == QuotedPrint)
        sup_quoted ();

    // et de l'encodage base64
    else if (encodage_texte == Base64)
    {
        i = decode64 (buf_lect);
        buf_lect [i++] = '\n';
        buf_lect [i] = '\0';
    }

    // sinon, on rajoute juste un passage  la ligne
    else
    {
        i = strlen (buf_lect);
        buf_lect [i++] = '\n';

        // terminaison de la chaine de caractres
        buf_lect [i] = '\0';
    }

    // convertir les caractres HTML, par exemple les &#8217; en apostrophes
    // ce genre de squence NE DEVRAIT JAMAIS se trouver dans un TEXTE de
    // mail correctement gnr (mais on en trouve !!!!)
    conv_carhtm ('#');

    // conversion des jeux de caractres non ISO
    // traitement des jeux de caractres

    // si mail encod UTF-8
    if (charset_texte == Utf8)
    {
        // et qu'on utilise le jeu de caractres ISO-8859
        if (! util_utf8 ())
            // conversion du jeu de caractres UTF8 en ISO-8859
            conv_utf8_iso ();
    }
    // sinon le mail est suppos encod ISO-8859
    else
    {
        // si on utilise le jeu de caractres UTF-8
        if (util_utf8 ())
            // conversion du jeu de caractres ISO-8859 en UTF8
            conv_iso_utf8 (buf_lect);   // fonction dans messages.c
        else
            // Normalement, le besoin de conversion des caractres
            // Mac devrait tre signale dans le charset.
            // Toutefois, il est prfrable faire la conversion systmatiquement
            conv_isomac ();
    }
}



/* suppression de l'encodage quoted-printable */

void sup_quoted ()
{
    int carcode;  // caractre correspondant  un codage quoted-printable
    int i, j;     // compteurs


    // initialisation
    i = 0;
    j = 0;

    // exploration de toute la ligne
    while (buf_lect [i] != '\0')
    {
        // si caractre quoted printable  convertir
        if (buf_lect [i] == '=' &&
             ((buf_lect [i+1] >= '0' && buf_lect [i+1] <= '9') ||
              (buf_lect [i+1] >= 'A' && buf_lect [i+1] <= 'F')))
        {
            // on dcode le caractre correspondant
            carcode = hexa (buf_lect + ++i);

            // si ce n'est pas un caractre de controle ou si c'est un LF
            if (carcode > 0x1F || carcode == '\n')
                // le mmoriser
                // Remarque : la conversion d'un LF peut tre une source de
                //            problme d'affichage notamment avec vmailfic
                buf_lect [j++] = carcode;

            // sinon, si caractre de controle autre que CR
            else if (carcode != '\r')
                // mmoriser un blanc pour limiter les problmes d'affichage
                buf_lect [j++] = ' ';

            // se positionner sur le caractre suivant
            i = i + 2;
        }
        // sinon on conserve le caractre courant et on passe au suivant
        else
            buf_lect [j++] = buf_lect [i++];
    }

    // traitement de la fusion ventuelle de 2 lignes
    if (buf_lect [i - 1] == '=')
        j--;
    else
        buf_lect [j++] = '\n';

    // terminaison de la chaine de caractres
    buf_lect [j] = '\0';
}



/* conversion des caractres spciaux sur Mac */

void conv_isomac ()
{
    int i;   // compteur


    for (i = 0; buf_lect [i] != '\0'; i++)
    {
        switch (buf_lect [i])
        {
            case 0x92 :
            case 0xB4 : buf_lect [i] = '\'';
                        break;
            case 0x93 : buf_lect [i] = '';
                        break;
            case 0x94 : buf_lect [i] = '';
                        break;
            case 0x9C : buf_lect [i] = '';
            default   : break;
        }
    }
}



/* conversion du jeu de caractres UTF8 en ISO-8859 */

void conv_utf8_iso ()
{
    /* indique que le caractre qui suit le  devra tre converti
       on utilise une variable statique pour garder l'info en mmoire dans le
       cas o le  et le caractre qui suit seraient sur 2 lignes distinctes */
    static int decalsuiv = 0;

    /* idem si  , suivi du caractre 0x80, suivi d'un 3me */
    static int codespe = 0;

    int i, j;   // compteurs


    // initialisation
    i = 0;
    j = 0;

    do
    {
        // si aucun traitement d'encodage sur 3 caractres en cours
        if (!codespe)
        {
            switch (buf_lect [i])
            {
                // si caractre  , on convertira le caractre qui le suit
                case 0xC3 : decalsuiv = 1;
                case 0xC2 : break; // on saute le caractre  ou 

                // si caractre  suivi du caractre 93h
                case 0xC5 : if (buf_lect [i+1] == 0x93)
                            {
                                // on convertit le rsultat en  
                                buf_lect [j++] = 0xBD;
                                i++;
                            }
                            else
                                buf_lect [j++] = buf_lect [i];

                            break;


                // si caractre  , mmorisation encodage sur 3 caractres
                case 0xE2 : codespe = 1;
                            break;

                // pour tous les autres caractres, 2 cas possibles
                default   : if (decalsuiv)
                            {
                                // conversion du caractre qui suit le 
                                buf_lect [j++] = buf_lect [i] | 0x40;
                                decalsuiv = 0;
                            }
                            else
                                // ou simple copie du caractre courant
                                buf_lect [j++] = buf_lect [i];
            }
        }
        // sinon, si on est sur le 2me caractre d'un encodage sur 3 car
        else if (codespe == 1)
        {
            // cas gnral (le 3me caractre dterminera le rsultat)
            if (buf_lect [i] == 0x80)
                codespe ++;

            // cas particulier du symbole euro
            else if (buf_lect [i] == 0x82 && buf_lect [i+1] == 0xAC)
            {
                buf_lect [j++] = 0xA4;
                i++;
                codespe = 0;
            }

            // encodage non reconnu
            else
            {
                buf_lect [j++] = 0xE2;
                buf_lect [j++] = buf_lect [i];
                codespe = 0;
            }
        }
        // traitement du dernier caractre d'un encodage sur 3 car
        else if (codespe == 2)
        {
            switch (buf_lect [i])
            {
                case 0x99: buf_lect [j++] = '\'';
                           break;

                case 0x93:
                case 0x9C:
                case 0x9D:
                case 0x9E: buf_lect [j++] = '"';
                           break;

                           // pointills
                case 0xA6: buf_lect [j++] = '.';

                           // on ne les met que si c'est possible sans
                           // craser les caractres non encore traits
                           if (i > j)
                               buf_lect [j++] = '.';

                           if (i > j)
                               buf_lect [j++] = '.';

                           break;

                // cas d'une squence encode non rpertorie
                default  : if (i > j + 1)
                           {
                               // mme prcaution que pour pointills
                               buf_lect [j++] = 0xE2;
                               buf_lect [j++] = 0x80;
                           }

                           buf_lect [j++] = buf_lect [i];
            }

            codespe = 0;
        }

        // si fin de ligne, on ne remet pas  jour decalsuiv et codespe
        if (buf_lect [++i] == 0)
            buf_lect [j++] = '\0';
    }
    while (buf_lect [i] != 0);
}



/* convertir les caractres HTML, par exemple les &#8217; en apostrophes */

void sup_balhtm ()
{
    static int dansbal = 0;
    int i, j;


    // initialisation
    i = 0;
    j = 0;

    // pour chaque caractre de la ligne
    while (buf_lect [i])
    {
        // si caractre <
        if (buf_lect [i] == '<')
            // on rentre dans une balise HTML
            dansbal = 1;

        // si on n'est pas dans un balise HTML
        if (! dansbal)
            // on affichera le caractre courant
            buf_lect [j++] = buf_lect [i];

        // si caractre >
        if (buf_lect [i] == '>')
            // on sort d'une balise HTML
            dansbal = 0;

        // passer au caractre suivant
        i++;
    }
    while (buf_lect [i++]);

    // terminer la ligne si ncessaire
    buf_lect [j] = '\0';
}



/* convertir les caractres HTML, par exemple les &#8217; en apostrophes */

void conv_carhtm (char limitation)
{
    // caractres auxquels est associ un nombre prcd d'un # ou un nom
    char *carnomme [] = {"#339", "-#8211", "'#8216", "'#8217", " nbsp",
                         "<lt", ">gt", "&amp", "'quot", "euro", "agrave",
                         "eacute", NULL};

    int trouve;  // indique si on a trouv un caractre nomm
    int choix;   // numro du caractre nomm en cours de test
    int depl;    // numro du caractre nomm test
    int i, j;    // position dans buf_lect en lecture et en criture


    // initialisation
    i = 0;
    j = 0;

    do
    {
        // si on trouve les caractres & et on traite tous les caractre nomms
        // ou si on trouve les caractres &#  la suite
        if (buf_lect [i] == '&' &&
                (buf_lect [i+1] == limitation || !limitation))
        {
            // initialisation
            trouve = 0;
            choix = 0;

            // chercher un ventuel carractre nomm
            while (!trouve && carnomme [choix])
            {
                // on vrifie caractre par caractre
                depl = 1;

                while (tolower (buf_lect [i + depl]) == carnomme [choix][depl])
                    depl ++;

                // si caractre nomm trouv
                if (buf_lect [i + depl] == ';' && !carnomme [choix][depl])
                    // mmoriser cette information
                    trouve = 1;
                else
                    // sinon, essayer avec le caractre nomm suivant
                    choix ++;
            }

            // si caractre nomm trouv
            if (trouve)
            {
                // le remplacer par le caractre adquat
                buf_lect [j++] = carnomme [choix][0];

                // et sauter la taille du caractre nomm
                i = i + depl;
            }
            // sinon
            else
                // simple copie du caractre courant
                buf_lect [j++] = '&';
        }
        // sinon
        else
            // simple copie du caractre courant
            buf_lect [j++] = buf_lect [i];
    }
    // on fait ce traitement jusqu' la fin de la ligne
    while (buf_lect [i++] != 0);
}



/* retourne la position d'un nom de fichier joint dans la ligne courante */

int posnomfic ()
{
    char chaineref [] = "name="; // chaine  chercher pour trouver fichier joint
    int i, j;                    // compteurs


    if (strlen (buf_lect) < 7)
        // la ligne est trop courte pour contenir un nom de fichier joint
        return (0);

    // le mot cl name ou filename ne commence pas en premire colonne
    i = 1;

    // recherche de la chaine name =
    do
    {
        // test sur le premier caractre
        if (tolower (buf_lect [i]) == *chaineref)
        {
            // et les suivants
            j = 1;

            while (tolower (buf_lect [i+j]) == chaineref [j])
                j++;

            // trouv ?
            if (chaineref [j] == '\0')
            {
                // sauter le " ventuel
                if (buf_lect [i+j] == '"')
                    i++;

                // et retourner la position du nom
                return (i + j);
            }
        }
    }
    // continuer la recherche
    while (buf_lect [++i] != '\0');

    // fichier joint non trouv
    return (0);
}
